@@ -1,5 +1,5 @@
 use serde::{de::DeserializeOwned, Serialize};
-use std::{collections::HashMap, fmt::Debug, path::Path};
+use std::{collections::HashMap, fmt::Debug, path::Path, sync::Arc};
 use tanoshi_lib::prelude::{Chapter, Extension, ExtensionResult, Filters, Manga, Param, Source};
 use tanoshi_util::http::Request;
 use tokio::{
@@ -26,15 +26,17 @@ impl ExtensionProxy {
     pub fn load<P: AsRef<Path>>(
         store: &Store,
         path: P,
-    ) -> Result<Box<dyn Extension>, Box<dyn std::error::Error>> {
+    ) -> Result<Arc<dyn Extension>, Box<dyn std::error::Error>> {
         let module = unsafe { Module::deserialize_from_file(&store, path)? };
 
-        let input = Pipe::new();
-        let output = Pipe::new();
+        let stdin = Pipe::new();
+        let stdout = Pipe::new();
+        let stderr = Pipe::new();
 
         let mut wasi_env = WasiState::new("tanoshi")
-            .stdin(Box::new(input))
-            .stdout(Box::new(output))
+            .stdin(Box::new(stdin))
+            .stdout(Box::new(stdout))
+            .stderr(Box::new(stderr))
             .finalize()?;
 
         let import_object = wasi_env.import_object(&module)?;
@@ -43,13 +45,18 @@ impl ExtensionProxy {
 
         let tanoshi = imports! {
             "tanoshi" => {
-                "host_http_request" => Function::new_native_with_env(&store, env.clone(), host_http_request)
+                "host_http_request" => Function::new_native_with_env(&store, env.clone(), host_http_request),
+                "host_debug" => Function::new_native_with_env(&store, env.clone(), host_debug),
+                "host_error" => Function::new_native_with_env(&store, env.clone(), host_error),
+                "host_info" => Function::new_native_with_env(&store, env.clone(), host_info),
+                "host_trace" => Function::new_native_with_env(&store, env.clone(), host_trace),
+                "host_warn" => Function::new_native_with_env(&store, env.clone(), host_warn),
             }
         };
 
         let instance = Instance::new(&module, &tanoshi.chain_back(import_object))?;
 
-        Ok(Box::new(ExtensionProxy { instance, env }))
+        Ok(Arc::new(ExtensionProxy { instance, env }))
     }
 
     #[cfg(not(feature = "disable-compiler"))]
@@ -237,7 +244,7 @@ pub async fn compile<P: AsRef<Path>>(path: P) -> Result<(), Box<dyn std::error::
 
 async fn thread(extension_receiver: UnboundedReceiver<Command>) {
     let mut recv = extension_receiver;
-    let mut extension_map: HashMap<i64, Box<dyn Extension>> = HashMap::new();
+    let mut extension_map: HashMap<i64, Arc<dyn Extension>> = HashMap::new();
 
     let store = ExtensionProxy::init_store_headless();
 
@@ -296,23 +303,21 @@ async fn thread(extension_receiver: UnboundedReceiver<Command>) {
                 }
                 Command::GetMangaList(source_id, param, tx) => {
                     process(&extension_map, source_id, tx, |proxy| {
-                        proxy.get_manga_list(param.clone())
+                        proxy.get_manga_list(param)
                     });
                 }
                 Command::GetMangaInfo(source_id, path, tx) => {
                     process(&extension_map, source_id, tx, |proxy| {
-                        proxy.get_manga_info(path.clone())
+                        proxy.get_manga_info(path)
                     });
                 }
                 Command::GetChapters(source_id, path, tx) => {
                     process(&extension_map, source_id, tx, |proxy| {
-                        proxy.get_chapters(path.clone())
+                        proxy.get_chapters(path)
                     });
                 }
                 Command::GetPages(source_id, path, tx) => {
-                    process(&extension_map, source_id, tx, |proxy| {
-                        proxy.get_pages(path.clone())
-                    });
+                    process(&extension_map, source_id, tx, |proxy| proxy.get_pages(path));
                 }
             }
         }
@@ -320,19 +325,23 @@ async fn thread(extension_receiver: UnboundedReceiver<Command>) {
 }
 
 fn process<F, T>(
-    extension_map: &HashMap<i64, Box<dyn Extension>>,
+    extension_map: &HashMap<i64, Arc<dyn Extension>>,
     source_id: i64,
     tx: ExtensionResultSender<T>,
     f: F,
 ) where
-    F: Fn(&Box<dyn Extension>) -> ExtensionResult<T>,
+    F: FnOnce(Arc<dyn Extension>) -> ExtensionResult<T> + Send + 'static,
+    T: Send + 'static,
 {
     match extension_map.get(&source_id) {
         Some(proxy) => {
-            let res = f(proxy);
-            if tx.send(res).is_err() {
-                error!("receiver dropped");
-            }
+            let proxy = proxy.clone();
+            tokio::spawn(async move {
+                let res = f(proxy);
+                if tx.send(res).is_err() {
+                    error!("receiver dropped");
+                }
+            });
         }
         None => {
             error!("extension with id {} not found", source_id);
@@ -340,6 +349,14 @@ fn process<F, T>(
     }
 }
 
+fn wasi_read_err(env: &ExtensionEnv) -> Result<String, Box<dyn std::error::Error>> {
+    let mut state = env.wasi_env.state();
+    let wasm_stderr = state.fs.stderr_mut()?.as_mut().ok_or("no wasi sterr")?;
+    let mut buf = String::new();
+    wasm_stderr.read_to_string(&mut buf)?;
+    Ok(buf)
+}
+
 fn wasi_read(env: &ExtensionEnv) -> Result<String, Box<dyn std::error::Error>> {
     let mut state = env.wasi_env.state();
     let wasm_stdout = state.fs.stdout_mut()?.as_mut().ok_or("no wasi stdout")?;
@@ -387,3 +404,63 @@ fn host_http_request(env: &ExtensionEnv) {
         }
     }
 }
+
+fn host_debug(env: &ExtensionEnv) {
+    let message = match wasi_read_err(env) {
+        Ok(message) => message,
+        Err(e) => {
+            error!("error wasi_read: {}", e);
+            return;
+        }
+    };
+
+    tanoshi_util::log::print_debug(message);
+}
+
+fn host_error(env: &ExtensionEnv) {
+    let message = match wasi_read_err(env) {
+        Ok(message) => message,
+        Err(e) => {
+            error!("error wasi_read: {}", e);
+            return;
+        }
+    };
+
+    tanoshi_util::log::print_error(message);
+}
+
+fn host_info(env: &ExtensionEnv) {
+    let message = match wasi_read_err(env) {
+        Ok(message) => message,
+        Err(e) => {
+            error!("error wasi_read: {}", e);
+            return;
+        }
+    };
+
+    tanoshi_util::log::print_info(message);
+}
+
+fn host_trace(env: &ExtensionEnv) {
+    let message = match wasi_read_err(env) {
+        Ok(message) => message,
+        Err(e) => {
+            error!("error wasi_read: {}", e);
+            return;
+        }
+    };
+
+    tanoshi_util::log::print_trace(message);
+}
+
+fn host_warn(env: &ExtensionEnv) {
+    let message = match wasi_read_err(env) {
+        Ok(message) => message,
+        Err(e) => {
+            error!("error wasi_read: {}", e);
+            return;
+        }
+    };
+
+    tanoshi_util::log::print_warn(message);
+}
